(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   Evaluator2.sml
 *  o Description: .
 *)


structure Evaluator2 = struct

structure B = Tools
structure T = NuprlTerms
structure P = Primitive
structure E = EvalProperties

val empty_env = T.mk_nuprl_empty_env_term

fun lookup_binding env var =
    let val (opr,[([w],thunk),([],e)]) = T.dest_term env
    in if w = var
       then SOME (T.dest_pair 6 thunk)
       else lookup_binding e var
    end handle _ => NONE

structure VARS = BinarySetFn(type ord_key = string val compare = String.compare)

fun length_env env =
    let val (opr,[([w],thunk),([],e)]) = T.dest_term env
    in 1 + length_env e
    end handle _ => 0

fun domain env =
    let val (opr,[([w],thunk),([],e)]) = T.dest_term env
    in w :: domain e
    end handle _ => []

fun print_first_env env =
    let val (opr,[([w],thunk),([],e)]) = T.dest_term env
    in w ^ ":" ^ T.toStringTerm thunk
    end handle _ => ""

fun clean_env env vars =
    let val (opr,[([w],thunk),([],e)]) = T.dest_term env
    in if VARS.member (vars, w)
       then clean_env e vars
       else T.mk_nuprl_term opr [([w], thunk), ([], clean_env e vars)]
    end handle _ => env

fun add_env_bindings' env [] = env
  | add_env_bindings' env ((var, term, env') :: bindings) =
    if var = ""
    then add_env_bindings' env bindings
    else let fun mk_env (t, e) =
		 let val pair = T.mk_nuprl_pair_term t e
		 in T.mk_nuprl_cons_env_term var pair env
		 end
	     val p =
		 if T.is_nuprl_variable_term term
		 then case lookup_binding env' (T.dest_variable term) of
			  SOME p => p
			| NONE => (term, env')
		 else (term, env')
	 in add_env_bindings' (mk_env p) bindings
	 end

fun add_env_bindings env bindings =
    let val (vars, bindings') =
	    foldr (fn (b as (var, _, _), (vars, bindings)) =>
		      if var = ""
		      then (vars, bindings)
		      else (VARS.add (vars, var), b :: bindings))
		  (VARS.empty, [])
		  bindings
    in add_env_bindings' (clean_env env vars) bindings'
    end

fun closure2term' term env =
    let (*val _ = print "closure2term\n"*)
	(*val _ = raise Fail "closure2term\n"*)
	fun aux term bounds env =
	    if T.is_nuprl_iclosure_term term
	    then let val (t,e) = T.dest_iclosure term
		 in aux t bounds e
		 end
	    else if T.is_nuprl_variable_term term
	    then let val v = T.dest_variable term
		 in if VARS.member (bounds, v)
		    then term
		    else case lookup_binding env v of
			     SOME (t,e) => aux t bounds e
			   | NONE => term
		 end
	    else let val (opr,bterms) = T.dest_term term
		     fun addBounds vs = VARS.addList (bounds, vs)
		     val bterms' = map (fn (vs,t) => (vs, aux t (addBounds vs) env)) bterms
		 in T.mk_nuprl_term opr bterms'
		 end
    in aux term VARS.empty env
    end

fun closure2term term env =
    let fun aux term env =
	    if T.is_nuprl_iclosure_term term
	    then let val (t,e) = T.dest_iclosure term
		 in aux t e
		 end
	    else if T.is_nuprl_variable_term term
	    then let val v = T.dest_variable term
		 in case lookup_binding env v of
			SOME (t,e) => aux t e
		      | NONE => term
		 end
	    else let val (opr,bterms) = T.dest_term term
		     val bterms' = map (fn (vs,t) => (vs, aux t (clean_env env (VARS.fromList vs)))) bterms
		 in T.mk_nuprl_term opr bterms'
		 end
    (*(\x.y)[y->(\u.x)[x->1]]*)
    in aux term env
    end

fun apply_ren ren (var : T.variable) =
    case List.find (fn (x,_) => x = var) ren of
	SOME (v,v') => v'
      | NONE => raise Fail "apply_ren"

fun accumulate2 f i [] [] = i
  | accumulate2 f i [] _  = raise Fail "accumulate2"
  | accumulate2 f i _  [] = raise Fail "accumulate2"
  | accumulate2 f i (x :: xs) (y :: ys) = accumulate2 f (f i x y) xs ys

fun all2 f [] [] = true
  | all2 f [] _  = false
  | all2 f _  [] = false
  | all2 f (x :: xs) (y :: ys) = (f x y) andalso (all2 f xs ys)

fun alpha_equal_closure_terms v1 e1 v2 e2 =
    let fun aux renamings (vs1,t1) env1 (vs2,t2) env2 =
	    let val renamings' = accumulate2 (fn r => fn u => fn v => (u,v) :: r) renamings vs1 vs2
		val ((op1,parms1),bts1) = T.dest_term t1
	    in if op1 = "!closure"
	       then let val (t, e) = T.dest_iclosure t1
		    in aux renamings' ([],t) e ([],t2) env2
		    end
	       else let val ((op2,parms2),bts2) = T.dest_term t2
		    in if op2 = "!closure"
		       then let val (t, e) = T.dest_iclosure t2
			    in aux renamings' ([],t1) env1 ([],t) e
			    end
		       else if op1 = "variable"
		       then let val v = T.dest_variable t1
			    in let val v' = apply_ren renamings' v
			       in op2 = "variable" andalso T.dest_variable t2 = v'
			       end handle _ => (case lookup_binding env1 v of
						    SOME (t,e) => aux renamings' ([],t) e ([],t2) env2
						  | NONE => raise Fail "alpha_equal_closure_terms")
			    end
		       else if op2 = "variable"
		       then let val v = T.dest_variable t2
			    in (op1 = "variable"
				andalso
				apply_ren renamings' (T.dest_variable t1) = v)
			       handle _ => (case lookup_binding env2 v of
						SOME (t,e) => aux renamings' ([],t1) env1 ([],t) e
					      | NONE => raise Fail "alpha_equal_closure_terms")
			    end
		       else op1 = op2
			    andalso all2 T.equal_parameters parms1 parms2
			    andalso all2 (fn x => fn y => aux renamings' x env1 y env2) bts1 bts2
		    end
	    end
    in aux [] ([],v1) e1 ([],v2) e2
    end handle _ => false

fun mct (term, env) =
    if T.alpha_equal_terms env empty_env
    then term
    else T.mk_nuprl_iclosure_term term env

fun decr x (n,s) = (x, (B.decr_steps n,s))
fun unit x s = (x,s)
fun exec f (t,e) s = (f s) handle _ => (print ("++failed(" ^ T.opid_of_term t ^ ")\n"); ((t,e), s))
fun >>= (m, k) =
 fn s =>
    let val (x,s') = m s
	(*val _ = (fn (steps,_) => print (">" ^ Int.toString steps ^ "\n")) s'*)
    in k x s'
    end

infix >>=

fun NextStepEval2 cls lib t principals non_principals env =
    let val opid = T.opid_of_term t
	(*val _ = print (opid ^ "\n")*)

    in if B.member opid ["add", "subtract", "multiply", "divide", "remainder"]
       then let val [(v1,e1),(v2,e2)] = principals
	    in decr (P.do_primitive_int_op opid v1 v2, empty_env, false)
	    end

       else if opid = "minus"
       then let val [(v,e)] = principals
	    in if T.is_nuprl_term "natural_number" v
	       then unit (T.mk_nuprl_minus_term v, empty_env, false)
	       else decr (P.do_primitive_minus v, env, false)
	    end

       else if B.member opid ["less", "int_eq"]
       then let val [(v1,e1),(v2,e2)] = principals
		val [([],t3),([],t4)] = non_principals
	    in if P.do_primitive_cmp opid v1 v2
	       then decr (t3, env, true)
	       else decr (t4, env, true)
	    end

       else if opid = "atom_eq"
       then let val n = (T.firstnat t) handle _ => 0
		val [(v1,e1),(v2,e2)] = principals
		val [([],t3),([],t4)] = non_principals
	    in if P.compare_atomn n v1 v2
	       then decr (t3, env, true)
	       else decr (t4, env, true)
	    end

       else if opid = "eq_term"
       then let val [(v1,e1),(v2,e2)] = principals
	    in if P.is_complete_primitive_value v1
		  andalso
		  P.is_complete_primitive_value v2
	       then if alpha_equal_closure_terms v1 e1 v2 e2
		    then decr (T.mk_nuprl_inl_term T.mk_nuprl_axiom_term, empty_env, false)
		    else decr (T.mk_nuprl_inr_term T.mk_nuprl_axiom_term, empty_env, false)
	       else raise Fail "eq_term"
	    end

       else if B.member opid ["isinr", "isinl", "ispair", "isint", "islambda", "isatom2"]
       then let val [(v1,e1)] = principals
		val [([],t2),([],t3)] = non_principals
	    in if P.do_primitive_test opid v1
	       then decr (t2, env, true)
	       else decr (t3, env, true)
	    end

       else if opid = "spread"
       then let val [(v1,e1)] = principals
		val [([x,y],B)] = non_principals
		val (a, b) = (T.dest_pair 5 v1)
		    handle _ => raise Fail "spread" (*(T.toStringTerm t)*)
	    in decr (B, add_env_bindings env [(y,b,e1),(x,a,e1)], true)
	    end

       else if opid = "decide"
       then let val [(v1,e1)] = principals
		val [([x],A),([y],B)] = non_principals
	    in if T.is_nuprl_term "inl" v1
	       then decr (A, add_env_bindings env [(x, T.subtermn 1 v1, e1)], true)
	       else if T.is_nuprl_term "inr" v1
	       then decr (B, add_env_bindings env [(y, T.subtermn 1 v1, e1)], true)
	       else raise Fail "decide"
	    end

       else if opid = "apply"
       then if List.null principals
	    then let val [([],yc),([],arg)] = non_principals
		 in if T.opid_of_term yc = "ycomb"
		    then decr (T.mk_nuprl_apply_term arg t, env, true)
		    else raise Fail "apply"
		 end
	    else let val [(f,fe)] = principals
		     val [([],arg)] = non_principals
		     (*val _ = print ("app:" ^ T.opid_of_term f ^ "," ^ T.opid_of_term arg ^ "\n")*)
		     val (x, B) = T.dest_lambda f
		 in decr (B, add_env_bindings fe [(x, arg, env)], true)
		 end

       else if opid = "callbyvalue"
       then let val [(q,e)] = principals
		val [([x],B)] = non_principals
	    in if P.is_primitive_value q
	       then decr (B, add_env_bindings env [(x, q, e)], true)
	       else raise Fail "callbyvalue"
	    end

       else if opid = "callbyvalueall"
       then let val [(q,e)] = principals
		val [([x],B)] = non_principals
	    in if P.is_complete_primitive_value q
	       then decr (B, add_env_bindings env [(x, q, e)], true)
	       else raise Fail ("callbyvalueall:(" ^ T.opid_of_term q ^ ")")
	    end

       else if opid = "list_ind"
       then let val [(q,e)] = principals
		val [([],nilcase),([x,xs,r],conscase)] = non_principals
	    in if T.is_nuprl_term "nil" q
	       then decr (nilcase, env, true)
	       else if T.is_nuprl_term "cons" q
	       then let val (qa,qb) = T.dest_cons q
			val qb' = mct (qb,e)
			val t2  = T.mk_nuprl_list_ind_term qb' nilcase (x,xs,r,conscase)
			val sub = [(x,qa,e),(xs,qb,e),(r,t2,env)]
		    in decr (conscase, add_env_bindings env sub, true)
		    end
	       else raise Fail "list_ind"
	    end

       else if opid = "ind"
       then let val [(q,e)] = principals
		val [([x,rd],downcase),([],basecase),([y,ru],upcase)] = non_principals
		val n = T.dest_integer q
		val (t',e') =
		    if n = 0 then (basecase,env)
		    else let val (m,r,w,c) =
				 if n > 0
				 then (n-1,ru,y,upcase)
				 else (n+1,rd,x,downcase)
			     val p  = T.mk_nuprl_integer_term m
			     val t2 = T.mk_nuprl_ind_term p (x,rd,downcase) basecase (y,ru,upcase)
			 in (c, add_env_bindings env [(w,q,e),(r,t2,env)])
			 end
	    in decr (t', e', true)
	    end

       else if opid = "rec_ind"
       then let val [([],arg),([f,x],B)] = non_principals
		val B' = T.mk_nuprl_rec_ind_term (T.mk_nuprl_variable_term x) (f,x,B)
		val function = T.mk_nuprl_lambda_term x B'
	    in decr (B, add_env_bindings env [(x,arg,env),(f,function,env)], true)
	    end

       else if opid = "variable"
       then case lookup_binding env (T.dest_variable t) of
		SOME (a,e) => unit (a, e, true)
	      | NONE => raise Fail ("variable " ^ T.toStringTerm t)

       else if opid = "!closure"
       then let val [([],u),([],e)] = non_principals
	    in unit (u, e, true)
	    end

       else if opid = "!abstraction"
       then let val [(v1,u1),(v2,u2),(v3,u3)] = principals
	    in unit (T.mk_nuprl_iabstraction_term v2 v3, env, false)
	    end

       else if opid = "!library"
       then let val [([],q),([],u)] = non_principals
	    in unit (q, env, true)
	    end

       else if E.is_termof_term lib t
       then decr (E.unfold_tof t, env, true)

       else if E.is_abstraction_term lib t
       then if cls
	    then decr (E.unfold_abs (SOME env) t, env, true)
	    else decr (E.unfold_abs NONE t, env, true)

       else if List.null (T.bterms_of_term t)
       then unit (t, empty_env, false)

       else unit (t, env, false)

    end

fun m_num_principals opid term (state as (steps,_)) =
    if steps < 0
       andalso opid = "apply"
       andalso T.is_nuprl_term "ycomb" (#1 (T.dest_apply term))
    then (0, state)
    else (E.num_principals opid, state)

fun clos_lst lst = map (fn c => ([], mct c)) lst

(* EVALUATOR2 - closure conversion *)
fun Evaluator2' cls lib term state =

    let fun EvalList cbva [] env = unit []
	  | EvalList cbva ((vars,t)::rest) env =
	    (Eval (t,env))
		>>=
		(fn p => ((if cbva then EvalAll else unit) p)
			     >>=
			     (fn q => (EvalList cbva rest env)
					  >>=
					  (fn vs => unit (q :: vs))))

	and Eval (t, env) =
	    let val (opr as (opid,params),subterms) = T.dest_term t
		(*val _ = (print (opid ^ "- "); app (fn (_,t) => print (" " ^ (if T.is_nuprl_variable_term t then T.toStringTerm t else T.opid_of_term t))) subterms; print "\n")*)
		(*val _ = if opid = "int_eq"
			then (* print ("\n----\n" ^ T.toStringTerm t ^ "\n----\n" ^ ListFormat.fmt {init = "[", sep = ",", final = "]", fmt = fn x => x} (domain env) ^ "\n-----\n")*)
			    print ("\n----\n" ^ T.toStringTerm t ^ "\n----\n" ^ print_first_env env ^ "\n-----\n")
			else ()*)
	    in (m_num_principals opid t)
		   >>=
		   (fn n =>
		       let val (p,np) = B.split n subterms
		       in (EvalList (E.is_eval_all opid) p env)
			      >>=
			      (fn c =>
			       fn state =>
				  let val ((t',e',ev),state') = NextStepEval2 cls lib t c np env state
				      (*val _ = print (opid ^ ":" ^ Bool.toString ev ^ "," ^ T.opid_of_term t' ^ "\n")*)
				  in if ev then Eval (t',e') state' else ((t',e'),state')
				  end handle _ => ((T.mk_nuprl_term opr ((clos_lst c) @ np), env), state))
		       end)
	    end

	(*and Eval (t, env) =
	    let val (opr as (opid,params),subterms) = T.dest_term t
		val _ = (print (opid ^ "- "); app (fn (_,t) => print (" " ^ (if T.is_nuprl_variable_term t then T.toStringTerm t else T.opid_of_term t))) subterms; print "\n")
		(*val _ = if opid = "int_eq"
			then (* print ("\n----\n" ^ T.toStringTerm t ^ "\n----\n" ^ ListFormat.fmt {init = "[", sep = ",", final = "]", fmt = fn x => x} (domain env) ^ "\n-----\n")*)
			    print ("\n----\n" ^ T.toStringTerm t ^ "\n----\n" ^ print_first_env env ^ "\n-----\n")
			else ()*)
	    in (m_num_principals opid t)
		   >>=
		   (fn n =>
		       let val (p,np) = B.split n subterms
		       in (EvalList (E.is_eval_all opid) p env)
			      >>=
			      (fn c => exec ((NextStepEval2 lib t c np env)
						 >>=
						 (fn (t',e',ev) => (print (opid ^ ":" ^ Bool.toString ev ^ "," ^ T.opid_of_term t' ^ "\n"); if ev then Eval else unit) (t',e')))
					    (T.mk_nuprl_term opr ((clos_lst c) @ np), env))
		       end)
	    end*)

	and EvalAll (t, env) =
	    let val (opr as (opid,params),subterms) = T.dest_term t
		(*val _ = (print (opid ^ "+ "); app (fn (_,t) => print (" " ^ T.opid_of_term t)) subterms; print "\n")*)
		val (principals,nprincipals) = B.split (E.num_principal_all opid) subterms
		val env' = if List.null nprincipals then empty_env else env
	    in (EvalList true principals env)
		   >>=
		   (fn c => unit (T.mk_nuprl_term opr ((clos_lst c) @ nprincipals), env'))
	    end

    in if T.is_nuprl_iclosure_term term
       then Eval (T.dest_iclosure term) state
       else Eval (term, empty_env) state
    end

fun Evaluator2 cls lib steps term =
    let val ((t',e'),(n,s)) = Evaluator2' cls lib term (steps, ())
	(*val _ = print (":" ^ Int.toString n ^ "," ^ T.opid_of_term t' ^  "\n")*)
	val answer =
	    if cls andalso T.is_nuprl_pair_term t'
	    then let val (s,msgs) = T.dest_pair 8 t'
		     val msgs' = closure2term msgs e'
		 in T.mk_nuprl_pair_term (T.mk_nuprl_iclosure_term s e') msgs'
		 end
	    else closure2term t' e'
    in (answer, steps - n)
    end

end
