(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        29 July 2011
 *  o File name:   NuprlTerms.sig
 *  o Description: Datatype for Nuprl terms.
 *)


signature NUPRL_TERMS = sig

    type variable        = string
    type opid            = string
    type tag             = string
    type parameter_kind  = string
    type parameter_value = string

    type opid_tag  = opid * tag
    type parameter = parameter_value * parameter_kind
    type operator  = opid_tag * parameter list

    datatype nuprl_term       = TERM   of operator * nuprl_bound_term list
	 and nuprl_bound_term = B_TERM of variable list * nuprl_term

    (* The signature of a term:
     *   - the types of the parameters
     *   - the int for each subterms *)
    type sign = (parameter_value option * parameter_kind) list * int list

    structure MAP : ORD_MAP where type Key.ord_key = string

    type item     = {id   : string,
		     sign : sign,
		     obid : string,
		     lhs  : nuprl_term,
		     rhs  : nuprl_term,
		     wfs  : nuprl_term list}
    type abs_lib  = item list MAP.map
    type tof_lib  = nuprl_term MAP.map
    type lib      = {abs : abs_lib, tof : tof_lib}

    val emlib      : lib
    val terms2map  : nuprl_term list -> lib
    val union_libs : lib -> lib -> lib
    val find_sign  : abs_lib -> nuprl_term -> item
    val insert_abs : lib -> string -> item -> lib
    val mk_item    : string          ->
		     sign            ->
		     string          ->
		     nuprl_term      ->
		     nuprl_term      ->
		     nuprl_term list ->
		     item

    val toStringTerm : nuprl_term -> string

    val ppTerm : nuprl_term -> string

    val nuprl2eml_wf  : string -> nuprl_term -> string
    val nuprl2eml_abs : string -> nuprl_term -> string
    val nuprlTerm2eml : nuprl_term -> string

    val opid_of_term : nuprl_term -> opid

    val mk_nuprl_obid_parameter        : parameter_value -> parameter
    val mk_nuprl_string_parameter      : parameter_value -> parameter
    val mk_nuprl_level_exp_parameter   : parameter_value -> parameter

    val mk_nuprl_ibool_term : bool -> nuprl_term

    val mk_nuprl_integer_term         : int -> nuprl_term
    val mk_nuprl_natural_number_term  : int -> nuprl_term

    val mk_nuprl_mkid_term  : string -> nuprl_term
    val mk_nuprl_itext_term : string -> nuprl_term

    val mk_nuprl_term : (opid * parameter list) -> (variable list * nuprl_term) list -> nuprl_term

    val mk_nuprl_simple_term : opid -> nuprl_term list -> nuprl_term

    val mk_nuprl_icomment_term : string -> nuprl_term -> nuprl_term

    val mk_nuprl_variable_term        : parameter_value -> nuprl_term
    val mk_nuprl_type_term            : parameter_value -> nuprl_term
    val mk_nuprl_msg_term             : parameter_value -> nuprl_term
    val mk_nuprl_event_orderingp_term : parameter_value -> nuprl_term
    val mk_nuprl_token_term           : parameter_value -> nuprl_term
    val mk_nuprl_istring_term         : parameter_value -> nuprl_term
    val mk_nuprl_valuealltype_term    : parameter_value -> nuprl_term
    val mk_nuprl_prop_term            : parameter_value -> nuprl_term

    val mk_nuprl_itheorem_term  : parameter_value -> nuprl_term -> nuprl_term
    val mk_nuprl_iproperty_term : parameter_value -> nuprl_term -> nuprl_term

    val mk_nuprl_iwf_lemmas_term : nuprl_term list -> nuprl_term

    val mk_nuprl_lambda_term : variable -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_term    : variable -> nuprl_term -> nuprl_term

    val mk_nuprl_loc_term               : nuprl_term
    val mk_nuprl_name_term              : nuprl_term
    val mk_nuprl_atom_term              : nuprl_term
    val mk_nuprl_real_term              : nuprl_term
    val mk_nuprl_unit_term              : nuprl_term
    val mk_nuprl_bool_term              : nuprl_term
    val mk_nuprl_int_term               : nuprl_term
    val mk_nuprl_loc_deq_term           : nuprl_term
    val mk_nuprl_bool_deq_term          : nuprl_term
    val mk_nuprl_atom_deq_term          : nuprl_term
    val mk_nuprl_nat_deq_term           : nuprl_term
    val mk_nuprl_unit_deq_term          : nuprl_term
    val mk_nuprl_int_deq_term           : nuprl_term
    val mk_nuprl_nil_term               : nuprl_term
    val mk_nuprl_icons_nil_term         : nuprl_term
    val mk_nuprl_itext_nil_term         : nuprl_term
    val mk_nuprl_it_term                : nuprl_term
    val mk_nuprl_nat_term               : nuprl_term
    val mk_nuprl_bfalse_term            : nuprl_term
    val mk_nuprl_btrue_term             : nuprl_term
    val mk_nuprl_empty_bag_term         : nuprl_term
    val mk_nuprl_event_ordering_p_term  : nuprl_term
    val mk_nuprl_axiom_term             : nuprl_term
    val mk_nuprl_ycomb_term             : nuprl_term
    val mk_nuprl_empty_env_term         : nuprl_term
    val mk_nuprl_inewline_term          : nuprl_term

    val mk_nuprl_list_term                    : nuprl_term -> nuprl_term
    val mk_nuprl_bag_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_deq_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_list_deq_term                : nuprl_term -> nuprl_term
    val mk_nuprl_lifting0_term                : nuprl_term -> nuprl_term
    val mk_nuprl_lifting1_term                : nuprl_term -> nuprl_term
    val mk_nuprl_lifting2_term                : nuprl_term -> nuprl_term
    val mk_nuprl_lifting3_term                : nuprl_term -> nuprl_term
    val mk_nuprl_lifting4_term                : nuprl_term -> nuprl_term
    val mk_nuprl_lifting_loc0_term            : nuprl_term -> nuprl_term
    val mk_nuprl_lifting_loc1_term            : nuprl_term -> nuprl_term
    val mk_nuprl_lifting_loc2_term            : nuprl_term -> nuprl_term
    val mk_nuprl_lifting_loc3_term            : nuprl_term -> nuprl_term
    val mk_nuprl_lifting_loc4_term            : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting0_term         : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting1_term         : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting2_term         : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting3_term         : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting4_term         : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting_loc0_term     : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting_loc1_term     : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting_loc2_term     : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting_loc3_term     : nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting_loc4_term     : nuprl_term -> nuprl_term
    val mk_nuprl_class_term                   : nuprl_term -> nuprl_term
    val mk_nuprl_bag_null_term                : nuprl_term -> nuprl_term
    val mk_nuprl_bag_only_term                : nuprl_term -> nuprl_term
    val mk_nuprl_bag_size_term                : nuprl_term -> nuprl_term
    val mk_nuprl_once_class_term              : nuprl_term -> nuprl_term
    val mk_nuprl_bnot_term                    : nuprl_term -> nuprl_term
    val mk_nuprl_pi1_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_pi2_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_send_once_class_term         : nuprl_term -> nuprl_term
    val mk_nuprl_send_once_loc_class_term     : nuprl_term -> nuprl_term
    val mk_nuprl_on_loc_class_term            : nuprl_term -> nuprl_term
    val mk_nuprl_primed_class_term            : nuprl_term -> nuprl_term
    val mk_nuprl_single_bag_term              : nuprl_term -> nuprl_term
    val mk_nuprl_msg_header_term              : nuprl_term -> nuprl_term
    val mk_nuprl_msg_type_term                : nuprl_term -> nuprl_term
    val mk_nuprl_msg_body_term                : nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined0_class_term     : nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined0_loc_class_term : nuprl_term -> nuprl_term
    val mk_nuprl_combined0_class_term         : nuprl_term -> nuprl_term
    val mk_nuprl_combined0_loc_class_term     : nuprl_term -> nuprl_term
    val mk_nuprl_inl_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_inr_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_isl_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_isr_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_outl_term                    : nuprl_term -> nuprl_term
    val mk_nuprl_outr_term                    : nuprl_term -> nuprl_term
    val mk_nuprl_evalall_term                 : nuprl_term -> nuprl_term
    val mk_nuprl_minus_term                   : nuprl_term -> nuprl_term
    val mk_nuprl_df_program_meaning_term      : nuprl_term -> nuprl_term
    val mk_nuprl_esE_term                     : nuprl_term -> nuprl_term
    val mk_nuprl_assert_term                  : nuprl_term -> nuprl_term

    val mk_nuprl_pair_term                        : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_apply_term                       : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_eq_atom_term                     : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_eq_bool_term                     : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_eq_int_term                      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_eq_loc_term                      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_prod_term                        : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_fun_term                         : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_union_term                       : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_bor_term                         : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_band_term                        : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_cons_term                        : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_proddeq_term                     : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_sumdeq_term                      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_combined_class_term              : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined_class_term          : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_combined_loc_class_term          : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined_loc_class_term      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_icons_cons_term                  : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_itext_cons_term                  : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_member_term                      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_primed_class_opt_term            : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_until_class_term                 : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_parallel_class_term              : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_programmable_term                : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_messages_delivered_term          : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_lt_int_term                      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_concat_term                      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_name_eq_term                     : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_divide_term                      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_multiply_term                    : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_subtract_term                    : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_add_term                         : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_iabstraction_term                : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_nlp_term                         : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_implies_term                     : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_uimplies_term                    : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_select_term                      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_general_base_class_term          : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_cons_bag_term                    : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_class_opt_term                   : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_class_opt_class_term             : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_lifting_gen_term                 : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_bag_map_term                     : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined1_class_term         : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_combined1_class_term             : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_combined1_loc_class_term         : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined0_class_opt_term     : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined1_loc_class_term     : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined0_loc_class_opt_term : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_iinclude_properties_term         : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_lifting_loc_gen_term             : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting_loc_gen_term      : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_concat_lifting_gen_term          : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_base_locs_headers_term           : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_class_at_term                    : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_base_headers_msg_val_term        : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_msg_has_type_term                : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_base_prc_term                    : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_once_prc_term                    : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_send_once_loc_prc_term           : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_on_loc_prc_term                  : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_prior_prc_term                   : nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_iclosure_term                    : nuprl_term -> nuprl_term -> nuprl_term

    val mk_nuprl_eq_bag_term                      : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_ite_term                         : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_comb_term                    : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_make_msg_term                    : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_std_ma_term                      : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_message_constraint_term          : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined2_class_term         : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_combined2_class_term             : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_combined2_loc_class_term         : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined1_class_opt_term     : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined2_loc_class_term     : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined1_loc_class_opt_term : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_base_headers_msg_val_loc_term    : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_base_at_prc_term                 : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_until_prc_term                   : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_at_prc_term                      : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_parallel_prc_term                : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_prior_init_prc_term              : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term

    val mk_nuprl_union_deq_term                    : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_product_deq_term                  : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined3_class_term          : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_combined3_class_term              : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_combined3_loc_class_term          : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined2_class_opt_term      : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined3_loc_class_term      : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined2_loc_class_opt_term  : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_bind_prc_term                     : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_loc_comb1_prc_term                : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined_loc_class1_prc_term  : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term

    val mk_nuprl_rec_combined3_class_opt_term      : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_rec_combined3_loc_class_opt_term  : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_classrel_term                     : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term

    val mk_nuprl_rec_comb_prc_term                 : nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term

    val mk_nuprl_spreadn_term     : nuprl_term -> (variable list * nuprl_term) -> nuprl_term

    val mk_nuprl_bind_class_term     : nuprl_term -> (variable * nuprl_term) -> nuprl_term
    val mk_nuprl_all_term            : nuprl_term -> (variable * nuprl_term) -> nuprl_term
    val mk_nuprl_uall_term           : nuprl_term -> (variable * nuprl_term) -> nuprl_term
    val mk_nuprl_isect_term          : nuprl_term -> (variable * nuprl_term) -> nuprl_term
    val mk_nuprl_callbyvalue_term    : nuprl_term -> (variable * nuprl_term) -> nuprl_term
    val mk_nuprl_callbyvalueall_term : nuprl_term -> (variable * nuprl_term) -> nuprl_term

    val mk_nuprl_list_ind_term : nuprl_term -> nuprl_term -> (variable * variable * variable * nuprl_term) -> nuprl_term

    val mk_nuprl_decide_term : nuprl_term -> (variable * nuprl_term) -> (variable * nuprl_term) -> nuprl_term

    val mk_nuprl_rec_ind_term : nuprl_term -> (variable * variable * nuprl_term) -> nuprl_term
    val mk_nuprl_spread_term  : nuprl_term -> (variable * variable * nuprl_term) -> nuprl_term

    val mk_nuprl_ind_term : nuprl_term -> (variable * variable * nuprl_term) -> nuprl_term -> (variable * variable * nuprl_term) -> nuprl_term

    val mk_nuprl_applies_term : nuprl_term -> nuprl_term list -> nuprl_term

    val mk_nuprl_let_term      : variable -> nuprl_term -> nuprl_term -> nuprl_term
    val mk_nuprl_cons_env_term : variable -> nuprl_term -> nuprl_term -> nuprl_term

    val mk_nuprl_genrec_term : variable * variable * nuprl_term -> nuprl_term

    val mk_nuprl_lambdas_term : variable list -> nuprl_term -> nuprl_term

    val mk_nuprl_finite_list_term : nuprl_term list -> nuprl_term

    val is_nuprl_term : opid -> nuprl_term -> bool

    val is_nuprl_variable_term              : nuprl_term -> bool
    val is_nuprl_lambda_term                : nuprl_term -> bool
    val is_nuprl_type_term                  : nuprl_term -> bool
    val is_nuprl_integer_term               : nuprl_term -> bool
    val is_nuprl_minus_natural_number_term  : nuprl_term -> bool
    val is_nuprl_iwf_lemmas_term            : nuprl_term -> bool
    val is_nuprl_iwftheorem_term            : nuprl_term -> bool
    val is_nuprl_iabstraction_term          : nuprl_term -> bool
    val is_nuprl_pair_term                  : nuprl_term -> bool
    val is_nuprl_cons_term                  : nuprl_term -> bool
    val is_nuprl_nil_term                   : nuprl_term -> bool
    val is_nuprl_select_term                : nuprl_term -> bool
    val is_nuprl_iclosure_term              : nuprl_term -> bool
    val is_nuprl_function_term              : nuprl_term -> bool
    val is_nuprl_bag_term                   : nuprl_term -> bool

    val dest_term         : nuprl_term -> (opid * parameter list) * (variable list * nuprl_term) list
    val dest_simple_term  : nuprl_term -> opid * nuprl_term list

    val dest_lambda       : nuprl_term -> variable * nuprl_term

    val dest_let          : nuprl_term ->  nuprl_term * variable * nuprl_term

    val dest_iabstraction : nuprl_term -> nuprl_term * nuprl_term * nuprl_term
    val dest_rec_comb     : nuprl_term -> nuprl_term * nuprl_term * nuprl_term

    val dest_variable     : nuprl_term -> variable

    val dest_so_variable  : nuprl_term -> variable * variable list

    val dest_integer      : nuprl_term -> int

    val dest_lambdas      : nuprl_term -> variable list * nuprl_term

    val dest_alls_ualls   : nuprl_term -> (variable * nuprl_term * bool) list * nuprl_term

    val dest_primed_class : nuprl_term -> nuprl_term
    val dest_minus        : nuprl_term -> nuprl_term
    val dest_iwftheorem   : nuprl_term -> nuprl_term
    val dest_bag          : nuprl_term -> nuprl_term

    val dest_product      : nuprl_term -> nuprl_term * nuprl_term
    val dest_function     : nuprl_term -> nuprl_term * nuprl_term
    val dest_cons         : nuprl_term -> nuprl_term * nuprl_term
    val dest_select       : nuprl_term -> nuprl_term * nuprl_term
    val dest_prior_prc    : nuprl_term -> nuprl_term * nuprl_term
    val dest_iclosure     : nuprl_term -> nuprl_term * nuprl_term
    val dest_apply        : nuprl_term -> nuprl_term * nuprl_term
    val dest_member       : nuprl_term -> nuprl_term * nuprl_term

    val dest_applies      : nuprl_term -> nuprl_term * nuprl_term list

    val dest_pair         : int -> nuprl_term -> nuprl_term * nuprl_term

    val dest_list         : nuprl_term -> nuprl_term list

    val dest_simple_null_term : nuprl_term -> opid * nuprl_term list

    val dest_eclass : nuprl_term -> nuprl_term * variable * variable * nuprl_term

    val fo_subst : (variable * nuprl_term) list -> nuprl_term -> nuprl_term

    val unfold_ab : nuprl_term option -> nuprl_term -> nuprl_term -> nuprl_term -> nuprl_term

    val toDeq : nuprl_term -> nuprl_term

    val subtermn : int -> nuprl_term -> nuprl_term

    val subterms : nuprl_term -> nuprl_term list

    val bterms_of_term : nuprl_term -> (variable list * nuprl_term) list

    val type_of_parameter  : parameter -> parameter_kind
    val value_of_parameter : parameter -> parameter_value

    val parameters_of_term : nuprl_term -> parameter list

    val equal_parameters : parameter -> parameter -> bool

    val alpha_equal_terms : nuprl_term -> nuprl_term -> bool

    val firstnat : nuprl_term -> int

    val get_obid_parameters : parameter list -> parameter_value option

    val getSignature : nuprl_term -> sign
    val eq_signs : sign -> sign -> bool

end
